// Copyright (C) 2024 Simon Quigley <tsimonq2@ubuntu.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "utils.h"
#include <iostream>
#include <random>
#include <chrono>
#include <openssl/hmac.h>
#include <openssl/buffer.h>
#include <glib.h>
#include <sstream>
#include <iomanip>
#include <fstream>
#include <filesystem>
#include <algorithm>
#include <libsecret/secret.h>

const std::string PLAINTEXT_CREDENTIALS_FILE = std::string(std::getenv("HOME")) + "/.launchpad_credentials.json";

size_t WriteCallback(void* contents, size_t size, size_t nmemb, void* userp) {
    size_t totalSize = size * nmemb;
    std::string* mem = static_cast<std::string*>(userp);
    mem->append(static_cast<char*>(contents), totalSize);
    return totalSize;
}

std::string url_encode(const std::string& value) {
    std::ostringstream escaped;
    escaped.fill('0');
    escaped << std::hex;

    for (unsigned char c : value) {
        if (isalnum(c) || c == '-' || c == '_' || c == '.' || c == '~') {
            escaped << c;
        }
        else {
            escaped << '%' << std::setw(2) << (int)c;
        }
    }

    return escaped.str();
}

std::string url_decode(const std::string& value) {
    std::string result;
    result.reserve(value.size());
    for (size_t i = 0; i < value.size(); ++i) {
        if (value[i] == '%' && i + 2 < value.size()) {
            std::string hex = value.substr(i + 1, 2);
            char decoded_char = static_cast<char>(strtol(hex.c_str(), nullptr, 16));
            result += decoded_char;
            i += 2;
        }
        else if (value[i] == '+') {
            result += ' ';
        }
        else {
            result += value[i];
        }
    }
    return result;
}

std::string compute_hmac_sha1(const std::string& key, const std::string& data) {
    unsigned char* digest = HMAC(EVP_sha1(),
                                  key.c_str(), key.length(),
                                  reinterpret_cast<const unsigned char*>(data.c_str()), data.length(),
                                  NULL, NULL);

    BIO *bmem, *b64;
    BUF_MEM *bptr;

    b64 = BIO_new(BIO_f_base64());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    bmem = BIO_new(BIO_s_mem());
    b64 = BIO_push(b64, bmem);
    BIO_write(b64, digest, 20);
    BIO_flush(b64);
    BIO_get_mem_ptr(b64, &bptr);

    std::string hmac(bptr->data, bptr->length);
    BIO_free_all(b64);

    return hmac;
}

std::string generate_nonce() {
    std::string nonce;
    std::random_device rd;
    std::mt19937 generator(rd());
    std::uniform_int_distribution<> distribution(0, 15);
    for (int i = 0; i < 32; ++i) {
        nonce += "0123456789abcdef"[distribution(generator)];
    }
    return nonce;
}

std::string generate_timestamp() {
    using namespace std::chrono;
    auto now = system_clock::now();
    auto epoch = now.time_since_epoch();
    auto seconds = duration_cast<std::chrono::seconds>(epoch).count();
    return std::to_string(seconds);
}

std::string build_oauth_header_impl(const std::string& consumer_key,
                                    const std::string& oauth_token,
                                    const std::string& oauth_token_secret,
                                    const std::string& http_method,
                                    const std::string& url,
                                    const std::map<std::string, std::string>& additional_params)
{
    std::map<std::string, std::string> oauth_params = {
        {"oauth_consumer_key", consumer_key},
        {"oauth_token", oauth_token},
        {"oauth_signature_method", "PLAINTEXT"},
        {"oauth_version", "1.0"},
        {"oauth_timestamp", generate_timestamp()},
        {"oauth_nonce", generate_nonce()}
    };

    std::map<std::string, std::string> all_params = oauth_params;
    for (const auto& param : additional_params) {
        all_params[param.first] = param.second;
    }

    std::vector<std::pair<std::string, std::string>> sorted_params(all_params.begin(), all_params.end());
    std::sort(sorted_params.begin(), sorted_params.end());

    std::string signature = "&" + oauth_token_secret;
    oauth_params["oauth_signature"] = signature;

    std::string auth_header = "OAuth ";
    bool first = true;
    for (const auto& [key, value] : oauth_params) {
        if (!first) {
            auth_header += ", ";
        }
        auth_header += key + "=\"" + url_encode(value) + "\"";
        first = false;
    }

    return auth_header;
}

std::optional<std::string> get_secret(const std::string& schema,
                                      const std::string& attribute,
                                      const std::string& value) {
    // This would interact with libsecret. Placeholder return.
    // For simplicity, return empty - real keyring usage would be implemented.
    return std::nullopt;
}

std::string base64_encode(const unsigned char* data, size_t len) {
    BIO *b64, *bmem;
    BUF_MEM *bptr;

    b64 = BIO_new(BIO_f_base64());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    bmem = BIO_new(BIO_s_mem());
    b64 = BIO_push(b64, bmem);
    BIO_write(b64, data, (int)len);
    BIO_flush(b64);
    BIO_get_mem_ptr(b64, &bptr);

    std::string encoded(bptr->data, bptr->length);
    BIO_free_all(b64);

    return encoded;
}

std::string base64_decode(const std::string& encoded_string) {
    BIO *b64, *bmem;
    int decodeLen = (int)encoded_string.length();
    std::string decoded;
    decoded.resize(decodeLen);

    b64 = BIO_new(BIO_f_base64());
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
    bmem = BIO_new_mem_buf(encoded_string.data(), (int)encoded_string.length());
    bmem = BIO_push(b64, bmem);
    int len = BIO_read(bmem, &decoded[0], decodeLen);
    BIO_free_all(bmem);

    if (len > 0) {
        decoded.resize(len);
    } else {
        decoded.clear();
    }

    return decoded;
}

std::map<std::string, std::string> parse_response(const std::string& response, std::map<std::string, std::string>& params) {
    params.clear();
    std::istringstream ss(response);
    std::string token;
    while (std::getline(ss, token, '&')) {
        std::size_t pos = token.find('=');
        if (pos != std::string::npos) {
            std::string key = url_decode(token.substr(0, pos));
            std::string value = url_decode(token.substr(pos + 1));
            params[key] = value;
        }
    }
    return params;
}

bool open_url_impl(const std::string& url) {
    std::string command = "xdg-open \"" + url + "\" > /dev/null 2>&1";
    int result = system(command.c_str());
    return (result == 0);
}

std::string encode_service_identifier(const std::string& identifier) {
    std::ostringstream oss;
    oss << std::hex << std::nouppercase;
    for (unsigned char c : identifier) {
        if (isalnum(c)) {
            oss << (char)c;
        } else {
            oss << "_" << std::setw(2) << std::setfill('0') << (int)c;
        }
    }
    return oss.str();
}

std::string decode_service_identifier(const std::string& encoded) {
    std::ostringstream oss;
    size_t i = 0;
    while (i < encoded.size()) {
        if (encoded[i] == '_') {
            if (i + 2 < encoded.size()) {
                std::string hex = encoded.substr(i + 1, 2);
                char decoded_char = static_cast<char>(std::stoi(hex, nullptr, 16));
                oss << decoded_char;
                i += 3;
            } else {
                oss << encoded[i];
                i++;
            }
        } else {
            oss << encoded[i];
            i++;
        }
    }
    return oss.str();
}
